# ***************************************************************
# Copyright (c) 2019 Dun Liang <randonlang@gmail.com>. All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
from jittor.nn import Pool, pool, AvgPool2d, avg_pool2d
from jittor.nn import MaxPool2d as j_MaxPool2d
from jittor.nn import max_pool2d as j_max_pool2d
import numpy as np
from .test_core import expect_error
from .test_grad import ngrad
from itertools import permutations
from jittor import compile_extern, Module
from .test_log import find_log_with_re
import random
import pickle as pk

skip_this_test = False

try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
    from torch.nn import MaxPool2d, Sequential
except:
    skip_this_test = True

class OldPool(Module):
    def __init__(self, kernel_size, stride=None, padding=0, dilation=None, return_indices=None, ceil_mode=False, count_include_pad=True, op="maximum"):
        assert dilation == None
        assert return_indices == None
        self.kernel_size = kernel_size
        self.op = op
        self.stride = stride if stride else kernel_size
        self.padding = padding
        self.ceil_mode = ceil_mode
        self.count_include_pad = count_include_pad and padding != 0

    def execute(self, x):
        N,C,H,W = x.shape
        if self.ceil_mode == False:
            h = (H+self.padding*2-self.kernel_size)//self.stride+1
            w = (W+self.padding*2-self.kernel_size)//self.stride+1
        else:
            h = (H+self.padding*2-self.kernel_size + self.stride - 1)//self.stride+1
            w = (W+self.padding*2-self.kernel_size + self.stride - 1)//self.stride+1

        # TODO: backward 
        xx = x.reindex([N,C,h,w,self.kernel_size,self.kernel_size], [
            "i0", # Nid
            "i1", # Cid
            f"i2*{self.stride}-{self.padding}+i4", # Hid
            f"i3*{self.stride}-{self.padding}+i5", # Wid
        ])
        return xx.reduce(self.op, [4,5])


def check(jt_model, torch_model, shape, near_data):
    if (near_data):
        assert shape[0] * shape[1] * shape[2] * shape[3] % 8 == 0
        data = list(range(8)) * int((shape[0] * shape[1] * shape[2] * shape[3]) / 8)
        random.shuffle(data)
        x = jt.array(data).float32().reshape(shape)
    else:
        x = jt.random(shape)
    y = jt_model(x)
    g = jt.grad(y.sum(), x)

    x_ = torch.Tensor(x.data)
    x_.requires_grad = True
    y_ = torch_model(x_)
    y_.sum().backward()
    y__ = y_.detach().numpy()
    g__ = x_.grad.detach().numpy()
    assert np.allclose(y.data, y__)
    assert np.allclose(g.data, g__)

@unittest.skipIf(skip_this_test, "No Torch found")
class TestArgPoolOp(unittest.TestCase):
    @unittest.skipIf(not jt.compiler.has_cuda, "No cuda found")
    @jt.flag_scope(use_cuda=1)
    def test_cuda(self):
        jt_model = jt.nn.Sequential(Pool(2, 2, 0), Pool(2, 2, 0), Pool(2, 2, 0, ceil_mode=True), Pool(2, 2, 0), Pool(2, 2, 0), Pool(3, 1, 1))
        torch_model = Sequential(MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0, ceil_mode=True), MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0), MaxPool2d(3, 1, 1))
        shape = [2, 3, 300, 300]
        check(jt_model, torch_model, shape, False)
        shape = [2, 3, 157, 300]
        check(jt_model, torch_model, shape, False)
        for i in range(10):
            check(jt_model, torch_model, [1,1,300,300], True)

    @unittest.skipIf(not jt.compiler.has_cuda, "No cuda found")
    @jt.flag_scope(use_cuda=1)
    def test_cuda_tuple(self):
        jt_model = jt.nn.Sequential(Pool((2,3), (2,3), (1,1)), Pool((2,3), (2,3), (1,1)), Pool((2,3), (2,3), (1,1), ceil_mode=True), Pool((2,3), (2,3), (1,1)), Pool((2,3), (2,3), (1,1)), Pool(3, 1, 1))
        torch_model = Sequential(MaxPool2d((2,3), (2,3), (1,1)), MaxPool2d((2,3), (2,3), (1,1)), MaxPool2d((2,3), (2,3), (1,1), ceil_mode=True), MaxPool2d((2,3), (2,3), (1,1)), MaxPool2d((2,3), (2,3), (1,1)), MaxPool2d(3, 1, 1))
        shape = [2, 3, 300, 300]
        check(jt_model, torch_model, shape, False)
        shape = [2, 3, 157, 300]
        check(jt_model, torch_model, shape, False)
        for i in range(10):
            check(jt_model, torch_model, [1,1,300,300], True)

    @unittest.skipIf(True, "TODO: cannot pass this test, fix me")
    @unittest.skipIf(not jt.compiler.has_cuda, "No cuda found")
    @jt.flag_scope(use_cuda=1)
    def test_cuda_old_pool(self):
        from torch.nn import AvgPool2d
        jt_model = OldPool(3, 1, 1, op="mean")
        torch_model = AvgPool2d(3, 1, 1)
        shape = [64, 64, 300, 300]
        check(jt_model, torch_model, shape, False)
        shape = [32, 128, 157, 300]
        check(jt_model, torch_model, shape, False)
        for i in range(10):
            check(jt_model, torch_model, [1,1,300,300], True)

    def test_cpu_(self):
        # x = jt.random([32, 128, 157, 300])
        x = jt.random([4, 128, 157, 300])
        x = jt.nn.pool(x, 2, "maximum", 0, 2)

    def test_cpu(self):
        jt_model = jt.nn.Sequential(Pool(2, 2, 0), Pool(2, 2, 0), Pool(2, 2, 0, ceil_mode=True), Pool(2, 2, 0), Pool(2, 2, 0), Pool(3, 1, 1))
        torch_model = Sequential(MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0, ceil_mode=True), MaxPool2d(2, 2, 0), MaxPool2d(2, 2, 0), MaxPool2d(3, 1, 1))
        # shape = [64, 64, 300, 300]
        shape = [4, 64, 300, 300]
        check(jt_model, torch_model, shape, False)
        # shape = [32, 128, 157, 300]
        shape = [4, 128, 157, 300]
        check(jt_model, torch_model, shape, False)
        for i in range(10):
            check(jt_model, torch_model, [1,1,300,300], True)
    
    def test_cpu_tuple(self):
        jt_model = jt.nn.Sequential(Pool((2,3), (2,3), (1,1)), Pool((2,3), (2,3), (1,1)), Pool((2,3), (2,3), (1,1), ceil_mode=True), Pool((2,3), (2,3), (1,1)), Pool((2,3), (2,3), (1,1)), Pool(3, 1, 1))
        torch_model = Sequential(MaxPool2d((2,3), (2,3), (1,1)), MaxPool2d((2,3), (2,3), (1,1)), MaxPool2d((2,3), (2,3), (1,1), ceil_mode=True), MaxPool2d((2,3), (2,3), (1,1)), MaxPool2d((2,3), (2,3), (1,1)), MaxPool2d(3, 1, 1))
        shape = [2, 3, 300, 300]
        check(jt_model, torch_model, shape, False)
        shape = [2, 3, 157, 300]
        check(jt_model, torch_model, shape, False)
        for i in range(10):
            check(jt_model, torch_model, [1,1,300,300], True)

    def test_index_pool(self):
        pool = jt.nn.Pool(2, return_indices=True)
        a = jt.randn([10,3,100,100])
        b, idx = pool(a)
        idx.sync()

    def test_index_pool2(self):
        pool = jt.nn.Pool(2, return_indices=True)
        a = jt.array([1,0,0,1,
                      0,0,0,0,
                      0,0,0,0,
                      1,0,0,1]).reshape((1,1,4,4))
        b, idx = pool(a)
        assert (idx.data.reshape((4,)) == [0,1,2,3]).all()

    def test_unpool(self):
        from jittor import nn
        pool = nn.MaxPool2d(2, stride=2, return_indices=True)
        unpool = nn.MaxUnpool2d(2, stride=2)
        input = jt.array([[[[ 1.,  2,  3,  4,0],
                                [ 5,  6,  7,  8,0],
                                [ 9, 10, 11, 12,0],
                                [13, 14, 15, 16,0],
                                [0,  0,  0,  0, 0]]]])
        output, indices = pool(input)
        out = unpool(output, indices, output_size=input.shape)
        assert (out == jt.array([[[[   0.,  0.,   0.,   0.,   0.],
                    [   0.,  6.,   0.,   8.,   0.],
                    [   0.,  0.,   0.,   0.,   0.],
                    [   0., 14.,   0.,  16.,   0.],
                    [   0.,  0.,   0.,   0.,   0.]]]])).all()

    def test_unpool_diff_kernel_stride(self):
        from jittor import nn
        pool = nn.MaxPool2d(3, stride=2, return_indices=True)
        unpool = nn.MaxUnpool2d(3, stride=2)
        input = jt.array([[[[ 1.,  2,  3,  4, 0],
                            [ 5,   6,  7,  8, 0],
                                [ 9, 10, 11, 12,0],
                                [13, 14, 16, 15,0],
                                [0,  0,  0,  0, 0]]]])
        output, indices = pool(input)
        out = unpool(output, indices, output_size=input.shape)
        assert (out == jt.array([[[
            [ 0.,  0.,  0.,  0.,  0.,],
            [ 0.,  0.,  0.,  0.,  0.,],
            [ 0.,  0., 11., 12.,  0.,],
            [ 0.,  0., 32.,  0.,  0.,],
            [ 0.,  0.,  0.,  0.,  0.,]]]])).all()

        

    @unittest.skipIf(not jt.compiler.has_cuda, "No cuda found")
    @jt.flag_scope(use_cuda=1)
    def test_cuda_avg_pool(self):
        self.test_cpu_avg_pool()

    def test_cpu_avg_pool(self):
        from torch.nn import AvgPool2d
        jt_model = Pool(2, 2, 0, op="mean", ceil_mode=True)
        torch_model = AvgPool2d(2, 2, 0, ceil_mode=True)
        shape = (2, 16, 33, 33)
        check(jt_model, torch_model, shape, False)

    def test_cpu_avg_pool2(self):
        from torch.nn import AvgPool2d
        jt_model = Pool(3, 1, 1, op="mean", ceil_mode=True)
        torch_model = AvgPool2d(3, 1, 1, ceil_mode=True)
        shape = (2, 16, 33, 33)
        check(jt_model, torch_model, shape, False)

    def test_AvgPool2d(self):
        from torch.nn import AvgPool2d as t_AvgPool2d
        jt_model = AvgPool2d(3, 1, 1, ceil_mode=True)
        torch_model = t_AvgPool2d(3, 1, 1, ceil_mode=True)
        shape = (2, 16, 33, 33)
        check(jt_model, torch_model, shape, False)

        jt_model = AvgPool2d(3, 1, 1, ceil_mode=True, count_include_pad=False)
        torch_model = t_AvgPool2d(3, 1, 1, ceil_mode=True, count_include_pad=False)
        shape = (2, 16, 100, 100)
        check(jt_model, torch_model, shape, False)
        print('finish')

    def test_avg_pool2d(self):
        from torch.nn.functional import avg_pool2d as t_avg_pool2d
        arr = np.random.random((2, 16, 33, 33))
        jt_model = avg_pool2d(jt.array(arr), 3, 1, 1, ceil_mode=True)
        torch_model = t_avg_pool2d(torch.Tensor(arr), 3, 1, 1, ceil_mode=True)
        assert np.allclose(jt_model.numpy(), torch_model.numpy())

        jt_model = avg_pool2d(jt.array(arr), 3, 1, 1, ceil_mode=True, count_include_pad=False)
        torch_model = t_avg_pool2d(torch.Tensor(arr), 3, 1, 1, ceil_mode=True, count_include_pad=False)
        assert np.allclose(jt_model.numpy(), torch_model.numpy())
        print('finish')

    def test_MaxPool2d(self):
        from torch.nn import MaxPool2d
        jt_model = j_MaxPool2d(3, 1, 1, ceil_mode=True)
        torch_model = MaxPool2d(3, 1, 1, ceil_mode=True)
        shape = (2, 16, 33, 33)
        check(jt_model, torch_model, shape, False)
        print('finish')

    def test_max_pool2d(self):
        from torch.nn.functional import max_pool2d
        arr = np.random.random((2, 16, 33, 33))
        jt_model = j_max_pool2d(jt.array(arr), 3, 1, 1, ceil_mode=True)
        torch_model = max_pool2d(torch.Tensor(arr), 3, 1, 1, ceil_mode=True)
        assert np.allclose(jt_model.numpy(), torch_model.numpy())

        jt_model = j_max_pool2d(jt.array(arr), 3, 1, 1)
        torch_model = max_pool2d(torch.Tensor(arr), 3, 1, 1)
        assert np.allclose(jt_model.numpy(), torch_model.numpy())

    def test_pool_3d(self):
        from torch.nn.functional import max_pool2d
        arr = np.random.random((2, 16, 20, 20, 20)).astype("float32")
        # arr = np.random.random((1, 1, 1, 5, 5)).astype("float32")
        jin = jt.array(arr)
        tin = torch.Tensor(arr)
        tin.requires_grad = True
        jt_model = jt.nn.Pool3d(3,1,1)(jin)
        torch_model = torch.nn.MaxPool3d(3,1,1)(tin)
        assert np.allclose(jt_model.numpy(), torch_model.detach().numpy())


        nout = np.random.random(tuple(jt_model.shape)).astype("float32")
        jout = jt_model * nout
        tout = torch_model * torch.Tensor(nout)
        dj = jt.grad(jout, jin)
        
        tout.sum().backward()
        dt = tin.grad
        assert np.allclose(dj.numpy(), dt.numpy())

    @unittest.skipIf(not jt.compiler.has_cuda, "No cuda found")
    @jt.flag_scope(use_cuda=1)
    def test_cuda_pool_3d(self):
        self.test_pool_3d()


if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import os
import jittor as jt
from .test_core import expect_error

header ="""
#pragma once
#include "op.h"

namespace jittor {

struct CustomOp : Op {
    Var* output;
    CustomOp(NanoVector shape, NanoString dtype=ns_float32);
    
    const char* name() const override { return "custom"; }
    DECLARE_jit_run;
};

} // jittor
"""

src = """
#include "var.h"
#include "custom_op.h"

namespace jittor {
#ifndef JIT
CustomOp::CustomOp(NanoVector shape, NanoString dtype) {
    output = create_output(shape, dtype);
}

void CustomOp::jit_prepare(JK& jk) {
    add_jit_define(jk, "T", output->dtype());
}

#else // JIT
#ifdef JIT_cpu
void CustomOp::jit_run() {
    index_t num = output->num;
    auto* __restrict__ x = output->ptr<T>();
    for (index_t i=0; i<num; i++)
        x[i] = (T)i;
}
#else
void CustomOp::jit_run() {
}
#endif // JIT_cpu
#endif // JIT

} // jittor
"""

class TestCustomOp(unittest.TestCase):
    def test_compile_custom_ops(self):
        tmp_path = jt.flags.cache_path
        hname = tmp_path+"/custom_op.h"
        ccname = tmp_path+"/custom_op.cc"
        with open(hname, "w") as f:
            f.write(header)
        with open(ccname, "w") as f:
            f.write(src)
        cops = jt.compile_custom_ops([hname, ccname])
        a = cops.custom([3,4,5], 'float')
        na = a.data
        assert a.shape == [3,4,5] and a.dtype == 'float'
        assert (na.flatten() == range(3*4*5)).all(), na

    def test_compile_custom_op(self):
        my_op = jt.compile_custom_op("""
        struct MyOp : Op {
            Var* output;
            MyOp(NanoVector shape, NanoString dtype=ns_float32);
            
            const char* name() const override { return "my"; }
            DECLARE_jit_run;
        };
        """, """
        #ifndef JIT
        MyOp::MyOp(NanoVector shape, NanoString dtype) {
            output = create_output(shape, dtype);
        }

        void MyOp::jit_prepare(JK& jk) {
            add_jit_define(jk, "T", output->dtype());
        }

        #else // JIT
        void MyOp::jit_run() {
            index_t num = output->num;
            auto* __restrict__ x = output->ptr<T>();
            for (index_t i=0; i<num; i++)
                x[i] = (T)-i;
        }
        #endif // JIT
        """,
        "my")
        a = my_op([3,4,5], 'float')
        na = a.data
        assert a.shape == [3,4,5] and a.dtype == 'float'
        assert (-na.flatten() == range(3*4*5)).all(), na

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers:
#     Guowei Yang <471184555@qq.com>
#     Dun Liang <randonlang@gmail.com>.
#
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

import jittor as jt
from jittor.optim import Optimizer
import math

class ReduceLROnPlateau(object):
    def __init__(self, optimizer, mode='min', factor=0.1, patience=10, verbose=False, threshold=1e-4, threshold_mode='rel', cooldown=0, min_lr=0, eps=1e-8):
        assert factor < 1.0, "factor should be < 1.0."
        assert isinstance(optimizer, Optimizer), '{} is not an Optimizer'.format(type(optimizer).__name__)
        assert mode in {'min', 'max'}, 'mode ' + mode + ' is unknown!'
        assert threshold_mode in {'rel', 'abs'},  'threshold mode ' + threshold_mode + ' is unknown!'

        if isinstance(min_lr, list) or isinstance(min_lr, tuple):
            assert len(min_lr) == len(optimizer.param_groups), "expected {} min_lrs, got {}".format(len(optimizer.param_groups), len(min_lr))
            self.min_lrs = list(min_lr)
        else:
            self.min_lrs = [min_lr] * len(optimizer.param_groups)
        self.factor = factor
        self.optimizer = optimizer
        self.patience = patience
        self.verbose = verbose
        self.cooldown = cooldown
        self.n_cd = 0
        self.mode = mode
        self.threshold = threshold
        self.threshold_mode = threshold_mode
        self.loss_best = None
        self.n_bad = 0
        self.eps = eps
        self.last_epoch = 0
        self.loss_best = math.inf if mode=="min" else -math.inf
        
    def step(self, loss, epoch=None):
        # convert `metrics` to float, in case it's a zero-dim Tensor
        loss_now = float(loss)
        if epoch is None:
            epoch = self.last_epoch + 1
        self.last_epoch = epoch

        if self.better(loss_now, self.loss_best):
            self.loss_best = loss_now
            self.n_bad = 0
        else:
            self.n_bad += 1

        if self.n_cd > 0:
            self.n_cd -= 1
            self.n_bad = 0

        if self.n_bad > self.patience:
            self.update_lr(epoch)
            self.n_cd = self.cooldown
            self.n_bad = 0
            
    def update_lr(self, epoch):
        for i, param_group in enumerate(self.optimizer.param_groups):
            old_lr = float(param_group.get("lr", self.optimizer.lr))
            new_lr = max(old_lr * self.factor, self.min_lrs[i])
            if old_lr - new_lr > self.eps:
                if param_group.get("lr")!=None:
                    param_group["lr"] = max(param_group["lr"] * self.factor, self.min_lrs[i])
                else:
                    self.optimizer.lr = new_lr
                if self.verbose:
                    print('Epoch {:5d}: reducing learning rate of group {} from {:.4e} to {:.4e}.'.format(epoch, i, old_lr, new_lr))
                          
    def better(self, a, b):
        if self.mode == 'min' and self.threshold_mode == 'rel':
            save = 1.0 - self.threshold
            return a < b * save
        elif self.mode == 'min' and self.threshold_mode == 'abs':
            return a < b - self.threshold
        elif self.mode == 'max' and self.threshold_mode == 'rel':
            save = self.threshold + 1.0
            return a > b * save
        else:
            return a > b + self.threshold

class CosineAnnealingLR(object):
    def __init__(self, optimizer, T_max, eta_min=0, last_epoch=-1):
        self.T_max = T_max
        self.eta_min = eta_min
        self.optimizer = optimizer
        self.last_epoch = last_epoch
        self.base_lr = optimizer.lr
        self.base_lr_pg = [pg.get("lr") for pg in optimizer.param_groups]
        #TODO set last_epoch is not ready

    def get_lr(self, base_lr, now_lr):
        if self.last_epoch == 0:
            return base_lr
        if (self.last_epoch - 1 - self.T_max) % (2 * self.T_max) == 0:
            return (now_lr + (base_lr - self.eta_min) *
                    (1 - math.cos(math.pi / self.T_max)) / 2)
        return  ((1 + math.cos(math.pi * self.last_epoch / self.T_max)) /
                (1 + math.cos(math.pi * (self.last_epoch - 1) / self.T_max)) *
                (now_lr - self.eta_min) + self.eta_min)

    def step(self):
        self.last_epoch += 1
        self.update_lr()
            
    def update_lr(self):
        self.optimizer.lr = self.get_lr(self.base_lr, self.optimizer.lr)
        for i, param_group in enumerate(self.optimizer.param_groups):
            if param_group.get("lr") != None:
                param_group["lr"] = self.get_lr(self.base_lr_pg[i], param_group["lr"])


class ExponentialLR(object):
    """ learning rate is multiplied by gamma in each step.
    """
    def __init__(self, optimizer, gamma, last_epoch=-1):
        self.optimizer = optimizer
        self.gamma = gamma
        self.last_epoch = last_epoch
        self.base_lr = optimizer.lr
        self.base_lr_pg = [pg.get("lr") for pg in optimizer.param_groups]

    def get_lr(self, base_lr, now_lr):
        if self.last_epoch == 0:
            return base_lr
        return base_lr * self.gamma ** self.last_epoch

    def step(self):
        self.last_epoch += 1
        self.update_lr()
            
    def update_lr(self):
        self.optimizer.lr = self.get_lr(self.base_lr, self.optimizer.lr)
        for i, param_group in enumerate(self.optimizer.param_groups):
            if param_group.get("lr") != None:
                param_group["lr"] = self.get_lr(self.base_lr_pg[i], param_group["lr"])


class StepLR(object):
    def __init__(self, optimizer, step_size, gamma=0.1, last_epoch=-1):
        self.optimizer = optimizer
        self.step_size = step_size
        self.gamma = gamma
        self.last_epoch = last_epoch
        self.cur_epoch = 0
    
    def get_gamma(self):
        if self.last_epoch < 0:
            if (self.cur_epoch != 0 and (self.cur_epoch + 1) % self.step_size == 0):
                return self.gamma
        else:
            if (self.cur_epoch + 1 + self.last_epoch) % self.step_size == 0:
                return self.gamma
        return 1.

    def get_lr(self):
        return self.optimizer.lr

    def step(self):
        self.update_lr()
        self.cur_epoch += 1
            
    def update_lr(self):
        gamma = self.get_gamma()
        if gamma != 1.:
            self.optimizer.lr = self.optimizer.lr * gamma
            for i, param_group in enumerate(self.optimizer.param_groups):
                if param_group.get("lr") != None:
                    param_group["lr"] = param_group["lr"] * gamma

class MultiStepLR(object):
    def __init__(self, optimizer, milestones=[], gamma=0.1, last_epoch=-1):
        self.optimizer = optimizer
        self.milestones = milestones
        self.gamma = gamma
        self.last_epoch = last_epoch
        #TODO set last_epoch is not ready
    
    def get_gamma(self):
        if (self.last_epoch in self.milestones):
            return self.gamma
        return 1.0

    def get_lr(self):
        now_lr = self.optimizer.lr
        return now_lr * self.get_gamma()

    def step(self):
        self.last_epoch += 1
        self.update_lr()
            
    def update_lr(self):
        gamma = self.get_gamma()
        if gamma != 1.0:
            self.optimizer.lr = self.optimizer.lr * gamma
            for i, param_group in enumerate(self.optimizer.param_groups):
                if param_group.get("lr") != None:
                    param_group["lr"] = param_group["lr"] * gamma



# ***************************************************************
# Copyright (c) 2021 Jittor.
# All Rights Reserved. 
# Maintainers:
#     Dun Liang <randonlang@gmail.com>.
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from jittor.test.test_grad import ngrad

class TestSlice(unittest.TestCase):
    def test_slice_bool(self):
        a = jt.zeros(10, "bool")
        a[1] = True
        a[2] = 1
        assert a.dtype == "bool"
        a.sync()
        assert np.equal(a.data, np.array([0,1,1,0,0,0,0,0,0,0])).all()

    def test_var_slices(self):
        def check(slices, msg):
            with jt.log_capture_scope() as logs:
                jt.core._print_var_slice(slices)
            s = logs[0]['msg']
            assert s == msg, s
        check((1), "[1,]")
        check(([[0],[1]],slice(None),[1,2],1), "[int32[2,1,],::,int32[2,],1,]")
        check((slice(None),slice(None),slice(None),slice(None)), "[::,::,::,::,]")
        check(([0,1],[0,1],[0,1],[0,1]), "[int32[2,],int32[2,],int32[2,],int32[2,],]")
        check(([0,1],-2,slice(None),[0,1]), "[int32[2,],-2,::,int32[2,],]")
        check(([0,1],slice(1,2,2),[1,2],1), "[int32[2,],1:2:2,int32[2,],1,]")
        check(([0,1],slice(None),[1,2],1), "[int32[2,],::,int32[2,],1,]")
        check((slice(1,None,2),slice(-1,None,2),[1,2],-4), "[1::2,-1::2,int32[2,],-4,]")
        check(0, "[0,]")
        check(10, "[10,]")
        check(-10, "[-10,]")
        check(1, "[1,]")
        check((1,slice(None),2), "[1,::,2,]")
        check((-2,slice(None),2,slice(1,9,2)), "[-2,::,2,1:9:2,]")
        check((None,1,None,2,None), "[-,1,-,2,-,]")
        check((...,1,...,2,...), "[...,1,...,2,...,]")

    @unittest.skipIf(not jt.has_cuda, "No cuda")
    @jt.flag_scope(use_cuda=1)
    def test_getitem(self):
        def check(shape, slices, i_to_vs="", i_to_o="", o_shape=""):
            # print(slices)
            x = jt.random(shape)

            with jt.log_capture_scope(log_vprefix="getitem=999") as logs:
                a = x.getitem(slices)
                a.sync()
            b = x.data[slices]
            bshape = b.shape if len(b.shape) else (1,)
            assert a.shape == bshape, (a.shape, bshape)
            s = logs[-1]['msg']
            assert "i_to_vs: "+i_to_vs in s
            assert "i_to_o: "+i_to_o in s
            assert "o_shape: "+o_shape in s
            aa = a.numpy()
            assert (aa==b).all(), (aa, b)

            y = x.numpy()
            v = jt.random(a.shape)
            z = x.setitem(slices, v)
            y[slices] = v.data
            assert (z.data==y).all(), (z.data, y, v.data, x.data)

            # test_setitem broadcast
            adim = len(a.shape)
            for mask in range(1<<adim):
                new_shape = list(a.shape)
                for i in range(adim):
                    if (mask>>i)&1:
                        new_shape[i] = 1
                y = x.numpy()
                v = jt.random(new_shape)
                z = x.setitem(slices, v)
                y[slices] = v.data
                assert (z.data==y).all(), (z.data, y, v.data, x.data)
                

        # TODO: when slice same row/col many times and assign value, numpy will retain the last value but we assign their sum. eg: check([3,3,3,3], ([[0,1,1]],slice(None),[[1],[2],[0]],1))
        check([3], (1), "[0,]", "[-1,]", "[]")
        check([3,3,3,3], ([[0],[1]],slice(None),[1,2],1), "[0,-1,2,3,]", "[-1,2,-1,-1,]", "[2,2,3,]")
        check([3,3,3,3], (slice(None),slice(None),slice(None),slice(None)), "[-1,-2,-2,-2,]", "[0,0,0,0,]", "[81,]")
        check([3,3,3,3], ([0,1],[0,1],[0,1],[0,1]), "[0,1,2,3,]", "[-1,-1,-1,-1,]", "[2,]")
        check([3,3,3,3], ([0,1],-2,slice(None),[0,1]), "[0,1,-1,3,]", "[-1,-1,1,-1,]", "[2,3,]")
        check([3,3,3,3], ([0,1],slice(1,2,2),[1,2],1), "[0,1,2,3,]", "[-1,1,-1,-1,]", "[2,1,]")
        check([3,3,3,3], ([0,1],slice(None),[1,2],1), "[0,-1,2,3,]", "[-1,1,-1,-1,]", "[2,3,]")
        check([3,3,3,3], (slice(1,10,1),...,slice(2,None,-1)), "[0,-1,-2,2,]", "[0,1,1,2,]", "[2,9,3,]")
        check([10,10,10,10], (slice(1,None,2),slice(-1,None,2),[1,2],-4), "[0,1,2,3,]", "[0,1,-1,-1,]", "")
        check([20], 0, "[0,]", "[-1,]", "[]")
        check([20], 10, "[0,]", "[-1,]", "[]")
        check([20], -10, "[0,]", "[-1,]", "[]")
        check([10,10,10,10], 1, "[0,-1,-2,-2,]", "[-1,0,0,0,]", "[1000,]")
        check([10,10,10,10], (1,slice(None),2), "[0,-1,2,-1,]", "[-1,0,-1,1,]", "")
        check([10,10,10,10], (-2,slice(None),2,slice(1,9,2)), "[0,-1,2,3,]", "[-1,0,-1,1,]")

    def test_getitem_grad(self):
        shape = (10,)
        slices = slice(2,4)

        a = jt.random(shape)
        b = a.getitem(slices)
        mask = jt.random(b.shape)
        loss = b*mask
        da = jt.grad(loss, a)

        _, np_grad = ngrad(lambda vars: (vars[0][slices]*mask.data).sum(), [a.numpy()], 1e-3)
        assert np.allclose(da.numpy(), np_grad, atol = 1e-3), (da.numpy(), np_grad)

        shape = (10,)
        slices = slice(2,4)

        a = jt.random(shape)
        b = a.getitem(slices)
        b = jt.random(b.shape)
        c = a.setitem(slices, b)
        mask = jt.random(c.shape)
        loss = c*mask
        da,db = jt.grad(loss, [a,b])

        def numpy_grad(vars):
            a, b = vars
            a = a.copy()
            a[slices] = b
            return (a*mask.data).sum()

        _, (nda, ndb) = ngrad(numpy_grad, [a.numpy(), b.numpy()], 1e-3)
        assert np.allclose(da.numpy(), nda, atol = 1e-3)
        assert np.allclose(db.numpy(), ndb, atol = 1e-3)

    def test_vary_shape_setitem(self):
        a = jt.array([1,2,3,4,5])
        b = jt.array([1,2,3,4,5])
        c = tuple(jt.where(b>3))
        a[c] = 0
        assert (a.data == [1,2,3,0,0]).all()

    def test_numpy_scalar_slice(self):
        a = jt.random((2,2))
        b = np.array([1])[0]
        assert a[b].shape == [2]



if __name__ == "__main__":
    unittest.main()
